#include "main.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h> 
#include <fcntl.h>
AMI *AMI::list = NULL;

int AMI::init()
{
    fstream serverList;

    string ip,id,username,secret;
    int port;


    serverList.open(SERVER_CONFIG);
    while (serverList)
    {
	MSG *m = MSG::Read(serverList);
	if (m->complete)
	{
	    port = atoi(m->Attr("port"));

	    if (!port) port = 5038;
	    if (atoi(m->Attr("RetryPeriod"))==0)
		m->Attr("RetryPeriod") << toStr(60*5);
	    new AMI(m->Attr("Server"), m->Attr("Host"), port, m->Attr("Username"), m->Attr("Secret"), atoi(m->Attr("RetryPeriod")));
	}
    }
    serverList.close();

    return 0;
}

void AMI::Shutdown()
{
    while (list)
	delete list;
}

AMI::~AMI()
{
    if (list==this) list = next;
    else
    {
	AMI *a = list;
	while (a && a->next && a->next != this) a = a->next;

	if (a) a->next = next;
    }

    delete MSG::Tie(this);
}

AMI::AMI(const char *_name, const char *_host, int _port, const char *_username, const char * _secret, time_t _retry)
{
    host = _host;
    port = _port;
    username = _username;
    secret = _secret;
    name = _name;

    errState=0;
    authenticated = false;
    control = 0;
    retryPeriod = _retry;

    m = NULL;

    next = NULL;

    AMI * a = list;
    if (!a) list = this;
    else
    {
	while (a->next) a = a->next;
	a->next = this;
    }
}

void AMI::loopAll()
{
    AMI *a, *a_next;

    for (a = list; a; a = a_next)
    {
	a_next = a->next;

	if (a->errState > time(NULL))
	    continue;

	if (!a->Loop())
	{
	    a->errState = time(NULL)+a->retryPeriod;
	    LOG::Warning(a->name+" AMI shut down for "+toStr(a->retryPeriod)+" seconds");
	}
    }
}


bool AMI::Loop()
{
    if (!Connected() && !Connect())
	return false;

    MSG::Read(this);

    if (initString == "") return true;

    if (!Authenticated())
    {
	if (!Authenticate())
	    return LOG::Error("Can't send authentication to "+name+" AMI"), Disconnect(), false;
	else
	    LOG::Notice("Sent authentication to "+name+" AMI");
    }

    return true;
}

bool AMI::Authenticate()
{
    if (authenticated)
	return true;

    (m = MSG::Tie(this))->clear();

    m->Pack("Action") << "Login";
    m->Pack("Username") << username;
    m->Pack("Secret") << secret;

    if (m->SendPack(&AMI::Authenticate_Reply))
	authenticated=1;
    return authenticated;
}

bool AMI::Connect()
{
    if ((control = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	return LOG::Error("Couldn't open socket"), false;

    struct hostent *host_resolved;
    struct sockaddr_in addr;
    host_resolved = gethostbyname(host.c_str());

    if (!host_resolved)
	return LOG::Error("Couldn't resolve host " + host + " to IP"), close(control), (control=false);

    memset((char*)&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    memcpy((void*)&addr.sin_addr.s_addr, host_resolved->h_addr, host_resolved->h_length);
    addr.sin_port = htons(port);

    if (connect(control, (struct sockaddr*)&addr, sizeof(addr)) < 0)
	return LOG::Error("Couldn't connect to host "+host+":"+toStr(port)), (control=false);

    if (fcntl(control, F_SETFL, fcntl(control, F_GETFL, 0)|O_NONBLOCK) < 0)
	LOG::Error("Couldn't make socket not block-read/write will be harmed");

    LOG::Notice("Connected to "+host+":"+toStr(port));
    return true;
}

void AMI::Disconnect()
{
    if (Connected())
    {
	close(control);
	control = 0;
    }

    authenticated = 0;
}


bool AMI::operator<<(string &msg)
{
    return (*this << msg.c_str());
}

bool AMI::operator<<(const char *msg)
{
    ssize_t c = strlen(msg);
    ssize_t x = write(control, msg, c);
    return (x==c);
}


void AMI::HandleMSG(MSG *msg)
{
    struct cb_list_type *cb,*cb_next,*cb_last=NULL;
    string actionID = msg->Attr("ActionID");
    string eventID = msg->Attr("Event");

    bool caught=false;

    if (actionID != "" || eventID != "")

    for (cb = cb_list; cb; cb = cb_next)
    {
	cb_next = cb->next;

	if ((cb->actionID != "" && cb->actionID == actionID) || (cb->eventID != "" && cb->eventID == eventID))
	{
	    if (cb->eventID == "")
	    {
		if (!cb_last) cb_list = cb->next;
		else cb_last->next = cb->next;
	    }

	    CALL_MEMBER_FN(*this, cb->func)(msg);

	    caught=true;

	    if (cb->eventID == "")
		delete cb;

	    continue;
	}
	cb_last =cb;
    }

    if (!caught && eventID != "Newexten" && 0)
    {
	cout << name << "\t" << eventID;
	if (eventID.length() <8) cout << "\t";
	MSGAttribute *a;int c = 0;
	for (a = msg->attr_list_in; a; a = a->next)
	{
	    if (a->attr != "Event" && a->value != "" && a->attr.find("Unique")==std::string::npos)
	    {
		cout << "\t" << a->attr << ": " << a->value;
		if (++c%2==0) cout << "\n\t\t\t";
	    }
	}
	cout << endl;
    }
}

void AMI::RegisterActionCB(string &actionID, AMI_CALLBACK cb)
{
    struct cb_list_type *cb_info = new cb_list_type;

    cb_info->actionID = actionID;
    cb_info->func = cb;
    cb_info->next = cb_list;
    cb_list=cb_info;
}

void AMI::RegisterEventCB(string &eventID, AMI_CALLBACK cb)
{
    struct cb_list_type *cb_info = new cb_list_type;

    cb_info->eventID = eventID;
    cb_info->func = cb;
    cb_info->next = cb_list;
    cb_list=cb_info;
}


void AMI::Authenticate_Reply(MSG *msg)
{
    string resp = msg->Attr("Response");
    if (authenticated==1 && resp == "Success")
    {
	LOG::Notice( name + " AMI Authentication response " + resp);
	authenticated=2;
	RegisterAllEvents();
    }
    else if (resp != "Success")
    {
	LOG::Error( name + " AMI Authentication response " + resp);
	Disconnect();
    }
}


struct AMI::cb_list_type AMI::event_callback_list[] =
{
    {"Dial", &AMI::__Dial}, 
    {"Newchannel", &AMI::__Newchannel},
    {"NewCallerid", &AMI::__NewCallerid},
    {"Newstate", &AMI::__Newstate},
    {"Bridge", &AMI::__Bridge},
    {"Link", &AMI::__Bridge},
    {"Unlink", &AMI::__Unlink},
    {"Hangup", &AMI::__Hangup},

};


void AMI::RegisterAllEvents()
{
    ssize_t len= sizeof(event_callback_list)/sizeof(cb_list_type);

    for (int x = 0; x < len; x++)
    {
	RegisterEventCB(event_callback_list[x].eventID, event_callback_list[x].func);
//	RegisterEventCB(event_callback_list[x].eventID, &AMI::HandleViaLog);
    }
}

void AMI::HandleViaLog(MSG *msg) 
{ 
    if (name != "apex-local") return; 

	cout <<name << "\t" << msg->Attr("Event");
	MSGAttribute *a;int c = 0;
	for (a = msg->attr_list_in; a; a = a->next)
	{
	    if (a->attr != "Event"/* && a->value != "" && a->attr.find("Unique")==std::string::npos*/)
	    {
		cout << "\t" << a->attr << ": " << a->value;
		if (++c%3==0) cout << "\n\t\t";
	    }
	}
	cout << endl;

}

void AMI::__Newchannel(MSG *msg)
{
    CHANNEL *chan = CHANNEL::Find(msg->Attr("Channel"));
    if (!chan) abort();
    chan->state = atoi(msg->Attr("ChannelState"));
    chan->context = msg->Attr("Context").value;
    chan->CIDNum = msg->Attr("CallerIDNum").value;
    chan->CIDName = msg->Attr("CallerIDName").value;

}

void AMI::__NewCallerid(MSG *msg)
{
    CHANNEL *ch = CHANNEL::Find(msg->Attr("Channel"));

    ch->CIDNum 	= msg->Attr("CallerIDNum").value;
    ch->CIDName = msg->Attr("CallerIDName").value;
}

void AMI::__Newstate(MSG *msg)	
{
    CHANNEL::Find(msg->Attr("Channel"))->state = atoi(msg->Attr("ChannelState"));
}

void AMI::__Dial(MSG *msg)
{
    string ch_str = msg->Attr("Channel").value;
    if (ch_str == "") ch_str = msg->Attr("Source").value;

    if (msg->Attr("SubEvent").value=="End")
    {
	CHANNEL *ch = CHANNEL::Find(ch_str, true);
	if (ch) delete ch;
    }
    else if (msg->Attr("SubEvent").value=="Begin" || msg->Attr("SubEvent").value=="")
    {
	LOG::Notice(ch_str + " --> "+msg->Attr("Destination").value + " "+msg->Attr("Dialstring").value);
    }
}

void AMI::__Hangup(MSG *msg)
{
    CHANNEL *ch = CHANNEL::Find(msg->Attr("Channel"), true);

    if (ch) delete ch;
}

void AMI::__Bridge(MSG *msg)
{
    CHANNEL *ch1 = CHANNEL::Find(msg->Attr("Channel1"), true),
	    *ch2 = CHANNEL::Find(msg->Attr("Channel2"), true);

    if (!ch1 || !ch2)
    {
//	LOG::Error("!ch1 or !ch2 in AMI::__Bridge() "+toStr(__LINE__)+(string)":"__FILE__);
	return;
    }

    if ((ch1->Bridge && ch1->Bridge != ch2)
	|| (ch2->Bridge && ch2->Bridge != ch1))
    {
//	LOG::Error("ch(n)->bridge && channel(n)->bridge != ch(x) in AMI::__Bridge() "+toStr(__LINE__)+(string)":"__FILE__);
	return;
    }

    LOG::Notice(ch1->name + " <-> "+ ch2->name);

    ch1->Bridge = ch2;
    ch2->Bridge = ch1;
}

void AMI::__Unlink(MSG *msg)	
{

    CHANNEL *ch1 = CHANNEL::Find(msg->Attr("Channel1"), true),
	    *ch2 = CHANNEL::Find(msg->Attr("Channel2"), true);

    if (!ch1 || !ch2)
    {
//	LOG::Error("!ch1 or !ch2 in AMI::__Unlink() "+toStr(__LINE__)+(string)":"__FILE__+" "+msg->Attr("Channel1").value+" ### "+msg->Attr("Channel2").value);
	return;
    }

    if ((ch1->Bridge && ch1->Bridge != ch2)
	|| (ch2->Bridge && ch2->Bridge != ch1))
    {
//	LOG::Error("ch(n)->bridge && channel(n)->bridge != ch(x) in AMI::__Unlink() "+toStr(__LINE__)+(string)":"__FILE__);
	return;
    }

    ch1->Bridge = NULL;
    ch2->Bridge = NULL;

    LOG::Notice(ch1->name + " ### "+ ch2->name);
}
