#include "main.h"
#include<sys/socket.h>
#include <errno.h>
#include <string.h>

MSG * MSG::list = NULL;

bool isspace2(char c)
{
   return (c == '\r' || c == '\t' || c == ' ');
}

MSG::MSG(iostream &stream)
{
    next = list;
    MSG::list = this;

    in = &stream;
    ami = NULL;

    complete = false;
    attr_list_in = attr_list_out = NULL;
}

MSG::~MSG()
{
    if (list == this)
	list = next;
    else
    {
	MSG *m = list;
	while (m && m->next && m->next != this) m = m->next;
	if (m) m->next = next;
    }

    if (attr_list_in)
	delete attr_list_in;
    if (attr_list_out)
	delete attr_list_out;

    if (initializedAMI) delete in;
}

MSG::MSG(AMI *_ami) 
{
    next = list;
    MSG::list = this;

    in = NULL;
    ami = _ami;
    initializedAMI=false;

    complete = false;
    attr_list_in = attr_list_out = NULL;
}

MSG *MSG::Tie(iostream &stream)
{
    MSG *msg;
    for (msg = list; msg; msg = msg->next)
	if (msg->in == &stream)
	    break;

    if (msg) 
	return msg;
    else
	return (new MSG(stream));;
}

MSG *MSG::Read(iostream &stream)
{
    return Tie(stream)->Read();
}

MSG *MSG::Tie(AMI *ami)
{
    MSG *msg;
    for (msg = list;msg; msg = msg->next)
	if (msg->ami == ami)
	    break;

    if (msg)
	return msg;
    else
	return new MSG(ami);
}

MSG *MSG::Read(AMI *ami)
{
    return Tie(ami)->ReadAMI();
}

MSG *MSG::ReadAMI()
{
    if (!ami->Connected())
	return this;

    while(1)
    {
	char buf[(MAX_LINE*2)+1];
	unsigned jump_len;
	if ((jump_len=unfinishedLine.length())>0)
	{
	    strcpy(buf, unfinishedLine.c_str());
	    unfinishedLine="";
	}
	ssize_t len = recv(ami->control, &buf[jump_len], (MAX_LINE*2)-jump_len, 0);

	if (len<0)
	{
	    if (errno != EAGAIN && errno != EWOULDBLOCK)
	    {
		LOG::Error("RECV LEN < 0 "+ami->name+" "+(string)strerror(errno));
		abort();
	    }

	    break;

	}
	if (len==0)
	{
	    LOG::Error("RECV LEN = 0 "+ami->name+" "+(string)strerror(errno));
	    LOG::Error("Disconnect not handled yet ("+ami->name+")");
	    abort();
	}

	len += jump_len;

	buf[len] = 0;

	int c=0,s;
	while (c < len)
	{
	    int y=0;
	    for (s=c; c < len && buf[c] != '\n'; c++)
	    {
		if (buf[c] == ':' && !y)
		{
		    y=c+1;
		    buf[c] = 0;
		}
	    }

	    if (c==len)
	    {
		buf[len] = 0;
		unfinishedLine = &buf[s];
		continue;
	    }

	    buf[c++]=0;

	    string left="", right="";

	    while (isspace2(buf[s]) && s < len) s++;
	    while (isspace2(buf[s+(strlen(&buf[s])-1)])) buf[s+(strlen(&buf[s])-1)]=0;
	    left = &buf[s];

	    while (y && isspace2(buf[y]) && y < len) y++;
	    while (isspace2(buf[y+(strlen(&buf[y])-1)]))  buf[y+(strlen(&buf[y])-1)]=0;
	    if (y) right = &buf[y];

	    if (!y)
		ami->initString = left;

	    if (left == "" && !y)
	    {
		ami->HandleMSG(this);
		delete attr_list_in;
		attr_list_in=NULL;
	    }
	    else
	    {
		Attr(left)<<right;
	    }
	}

    }
	
    return this;
}


MSG *MSG::Read()
{
    char c;

    if (ami)
	return ReadAMI();

    if (complete)
    {
	complete = false;
	delete attr_list_in;
	attr_list_in = NULL;
    }


    in->get(line_buf, MAX_LINE);
    in->get(c);

    if (c != '\n')
    {
	in->unget();
	unfinishedLine = unfinishedLine + line_buf;
    }
    if ((c == '\n'||!c) && line_buf[0]==0 && unfinishedLine.empty())
    {
	if (attr_list_in)
	{
	    in->clear();
	    complete = true;
	}

	while (isspace((c=in->get())));
	if (c>0) in->unget();
	return this;
    }

    if (c == '\n' || !in)
    {
	int x;
	string line;
	line = unfinishedLine + line_buf;
	unfinishedLine = "";

	if ((x=line.find(":"))>=0)
	{
	    string left = line.substr(0, x);
	    string right = line.substr(x+1);

	    while (left.length()>0 && left[0]==' ') 		left.erase(0,1);
	    while (right.length()>0 && right[0]==' ') 		right.erase(0,1);

	    while ((x=left.length())>0 && left[x-1]==' ') 	left.erase(x,1);
	    while ((x=right.length())>0 && right[x-1]==' ') 	right.erase(x,1);

	    Attr(left) << right;
	}
    }

    return this;
}

MSGAttribute &MSG::retr(MSGAttribute *& attr_list, string &left)
{
    MSGAttribute *a, *b;

    for (a = attr_list; a; a = a->next)
	if (a->attr == left) return *a;

    a = new MSGAttribute(left);

    if (!(b=attr_list)) attr_list = a;
    else
    {
	while (b->next) b = b->next;
	b->next = a;
    }

    return *a;
}

bool MSG::SendPack(AMI_CALLBACK cb)
{
    static unsigned long actionId=0;
    if (!attr_list_out || !ami->Connected()) return false; // nothing to send || not connected

    if(Pack("ActionID").value=="")
	Pack("ActionID") << "ACTION"+toStr(++actionId);

    if (cb)
	ami->RegisterActionCB(Pack("ActionID").value, cb);

    string msg = *this;

    return (*ami << msg);
}


